unit ListStuff;
{
Author: Craig Manley
Date: 10 January 2000
}

interface

uses Windows, Classes, SysUtils;

type
  {-- TContainerList --}
  TReferenceType = (rtStruct,rtClass);

  TContainerList = class(TList)
  protected
    FRefType: TReferenceType;
  public
    constructor Create(RefType: TReferenceType);
    destructor Destroy; override;
    procedure Clear;
    procedure Delete(Index: Integer);
    procedure Move(CurIndex, NewIndex: Integer);
    function Remove(Item: Pointer): Integer;
  end;

  {-- TValueList --}
  PPIdString = ^PIdString;
  PIdString = ^TIdString;
  TIdString = record
    ID: integer;
    Value: string;
  end;

  TValueList = class(TObject)
  private
    FBlockSize: WORD;
    FPointers: PPIdString;
    FCount: integer;
    FCapacity: integer;
    procedure CheckCapacity;
  public
    constructor Create(BlockSize: WORD);
    destructor Destroy; override;
    procedure Clear;
    function Delete(ID: integer): boolean;
    function Exists(ID: integer): boolean;
    function Insert(ID: integer; const Value: string): boolean;
    function Value(ID: integer): string;
  published
    property Count: Integer read FCount;
    property Capacity: Integer read FCapacity;
  end;

implementation

{-- TContainerList --}
constructor TContainerList.Create(RefType: TReferenceType);
begin
  FRefType := RefType;
  inherited Create;
end;

destructor TContainerList.Destroy;
begin
  Clear;
  inherited;
end;

procedure TContainerList.Clear;
var
  i: integer;
begin
  for i := Count-1 downto 0 do
    Delete(i);
  inherited;
end;

procedure TContainerList.Delete(Index: Integer);
begin
  if Items[Index] <> nil then
  begin
    if FRefType = rtClass then
      TObject(Items[Index]).Free
    else if FRefType = rtStruct then
      Dispose(Items[Index]);
  end;
  inherited;
end;

procedure TContainerList.Move(CurIndex, NewIndex: Integer);
begin
  if Items[NewIndex] <> nil then
  begin
    if FRefType = rtClass then
      TObject(Items[NewIndex]).Free
    else if FRefType = rtStruct then
      Dispose(Items[NewIndex]);
  end;
  inherited;
end;

function TContainerList.Remove(Item: Pointer): Integer;
type
  PPointer = ^pointer;
var
  pArray: pointer;
  p: PPointer; // p is a pointer to memory that contains a pointer. Confusing?
  i: integer;
begin
  GetMem(pArray,Count * SizeOf(Pointer));
  try
    p := pArray;
    for i := 0 to Count-1 do
    begin
      inc(p,i);
      p^ := Items[i];
    end;
    Result := inherited Remove(Item);
    if Result >= 0 then
    begin
      p := pArray;
      inc(p,Result);
      if p^ <> nil then
      begin
        if FRefType = rtClass then
          TObject(p^).Free
        else if FRefType = rtStruct then
          Dispose(p^);
      end;
    end;
  finally
    FreeMem(pArray);
  end;
end;


{-- TValueList --}
constructor TValueList.Create(BlockSize: WORD);
begin
  inherited Create;
  FBlockSize := BlockSize;
  if FBlockSize = 0 then
    inc(FBlockSize);
end;

destructor TValueList.Destroy;
begin
  Clear;
  ReallocMem(FPointers,0);
  inherited;
end;

procedure TValueList.CheckCapacity;
begin
  if FCount <= (FCapacity - FBlockSize) then
  begin
    if (FCount mod FBlockSize) > 0 then
      FCapacity := ((FCount div FBlockSize) + 1) * FBlockSize
    else
      FCapacity := FCount;
    ReallocMem(FPointers,FCapacity*SizeOf(PPIdString));
  end;
end;

procedure TValueList.Clear;
var
  p: PPIdString;
begin
  p := FPointers;
  while FCount > 0 do
  begin
    Dispose(p^);
    inc(p);
    dec(FCount);
  end;
  CheckCapacity;
end;

function TValueList.Delete(ID: Integer): boolean;
var
  i,j: integer;
  p: PPIdString;
begin
  p := FPointers;
  Result := False;
  for i := 0 to FCount-1 do
  begin
    if p^^.ID = ID then
    begin
      Dispose(p^);
      dec(FCount);
      for j := i to FCount-1 do
      begin
        p^ := PPIdString(DWORD(p)+SizeOf(PPIdString))^;
        inc(p);
      end;
      Result := True;
      Break;
    end;
    inc(p);
  end;
  CheckCapacity;
end;

function TValueList.Exists(ID: Integer): boolean;
var
  i: integer;
  p: PPIdString;
begin
  p := FPointers;
  Result := False;
  for i := 0 to FCount-1 do
  begin
    if p^^.ID = ID then
    begin
      Result := True;
      Break;
    end;
    inc(p);
  end;
end;

function TValueList.Insert(ID: integer; const Value: string): boolean;
var
  p: PPIdString;
begin
  Result := not Exists(ID);
  if Result then
  begin
    if FCount >= FCapacity then
    begin
      ReallocMem(FPointers,(FCapacity + FBlockSize)*SizeOf(PPIdString));
      inc(FCapacity,FBlockSize);
    end;
    p := FPointers;
    inc(p,FCount);
    new(p^);
    p^^.ID := ID;
    p^^.Value := Value;
    inc(FCount);
  end;
end;

function TValueList.Value(ID: integer): string;
var
  i: integer;
  p: PPIdString;
begin
  p := FPointers;
  Result := '';
  for i := 0 to FCount-1 do
  begin
    if p^^.ID = ID then
    begin
      Result := p^^.Value;
      Break;
    end;
    inc(p);
  end;
end;

end.
